import { Colors } from "./index.js";
import {
   Adapter,
   Message,
   TerminalFormats,
   Formatted,
   IFormatted,
   LoggingTypes,
} from "./types.js";

declare const Deno: any;
declare const process: any;

const browser = typeof window !== "undefined" && typeof Deno === "undefined";
const deno = typeof Deno !== "undefined";
const NodeJS = typeof process !== undefined;

export class ConsoleAdapter implements Adapter {
   level: LoggingTypes = LoggingTypes.Debug;
   constructor(private colors: boolean = true) {}

   init() {}
   flush() {}

   setLevel(level: LoggingTypes) {
      this.level = level;
   }
   // TODO: Check if required!
   // private escape(text: string): string {
   //    return text
   //       .replace(/%s/g, "%%s")
   //       .replace(/%c/g, "%%c")
   // }

   private format(
      formatted: IFormatted<string>[]
   ): [string, string[] | undefined] {
      let text = "";
      let style_formats: string[] = [];

      if (!browser) {
         // NodeJS or Deno
         for (const format of formatted) {
            let formats = "";
            if (format.bold) {
               formats += TerminalFormats.Bold;
            }

            if (format.blink) {
               formats += TerminalFormats.Blink;
            }

            if (format.underscore) {
               formats += TerminalFormats.Underscore;
            }

            if (format.color) {
               switch (format.color) {
                  case Colors.RED:
                     formats += TerminalFormats.FgRed;
                     break;
                  case Colors.GREEN:
                     formats += TerminalFormats.FgGreen;
                     break;
                  case Colors.YELLOW:
                     formats += TerminalFormats.FgYellow;
                     break;
                  case Colors.BLUE:
                     formats += TerminalFormats.FgBlue;
                     break;
                  case Colors.MAGENTA:
                     formats += TerminalFormats.FgMagenta;
                     break;
                  case Colors.CYAN:
                     formats += TerminalFormats.FgCyan;
                     break;
                  case Colors.WHITE:
                     formats += TerminalFormats.FgWhite;
                     break;
               }
            }

            if (format.bgcolor) {
               switch (format.bgcolor) {
                  case Colors.RED:
                     formats += TerminalFormats.BgRed;
                     break;
                  case Colors.GREEN:
                     formats += TerminalFormats.BgGreen;
                     break;
                  case Colors.YELLOW:
                     formats += TerminalFormats.BgYellow;
                     break;
                  case Colors.BLUE:
                     formats += TerminalFormats.BgBlue;
                     break;
                  case Colors.MAGENTA:
                     formats += TerminalFormats.BgMagenta;
                     break;
                  case Colors.CYAN:
                     formats += TerminalFormats.BgCyan;
                     break;
                  case Colors.WHITE:
                     formats += TerminalFormats.BgWhite;
                     break;
               }
            }

            text += formats + format.content + TerminalFormats.Reset;
            // (formats.length > 0 ? TerminalFormats.Reset : ""); //TODO: Benchmark if this is better
         }
      } else {
         for (const format of formatted) {
            let styles: string[] = [];
            let resetStyles: string[] = [];
            if (format.bold) {
               styles.push("font-weight: bold;");
               resetStyles.push("font-weight: unset");
            }

            if (format.underscore) {
               styles.push("text-decoration: underline");
               resetStyles.push("text-decoration: unset");
            }

            if (format.blink) {
               styles.push("text-decoration: blink");
               resetStyles.push("text-decoration: unset");
            }

            if (format.color) {
               let color = "";
               switch (format.color) {
                  case Colors.RED:
                     color = "red";
                     break;
                  case Colors.GREEN:
                     color = "green";
                     break;
                  case Colors.YELLOW:
                     color = "gold";
                     break;
                  case Colors.BLUE:
                     color = "blue";
                     break;
                  case Colors.MAGENTA:
                     color = "magenta";
                     break;
                  case Colors.CYAN:
                     color = "cyan";
                     break;
                  case Colors.WHITE:
                     color = "white";
                     break;
               }
               styles.push("color: " + color);
               resetStyles.push("color: unset");
            }

            if (format.bgcolor) {
               let color = "";
               switch (format.bgcolor) {
                  case Colors.RED:
                     color = "red";
                     break;
                  case Colors.GREEN:
                     color = "green";
                     break;
                  case Colors.YELLOW:
                     color = "gold";
                     break;
                  case Colors.BLUE:
                     color = "blue";
                     break;
                  case Colors.MAGENTA:
                     color = "magenta";
                     break;
                  case Colors.CYAN:
                     color = "cyan";
                     break;
                  case Colors.WHITE:
                     color = "white";
                     break;
               }
               styles.push("background-color: " + color);
               resetStyles.push("background-color: unset");
            }

            text += "%c" + format.content.replace(/%c/g, "%%c") + "%c";
            style_formats.push(styles.join(";"), resetStyles.join(";"));
         }
      }

      return [text, style_formats];
   }

   onMessage(message: Message) {
      if (browser) {
         if (this.colors) {
            const [text, formats] = this.format(message.text);
            console.log(text, ...formats);
         } else {
            console.log(Formatted.strip(message.text));
         }
      } else {
         const text = this.colors
            ? this.format(message.text)[0]
            : Formatted.strip(message.text);

         if (deno) {
            //TODO: Deno specific thing
            console.log(text);
         } else if (typeof process !== "undefined") {
            //NodeJS
            process.stdout.write(text + "\n");
         } else {
            console.log(text);
         }
      }
   }
}
