export enum LoggingTypes {
   Debug,
   Log,
   Warn,
   Error,
}

export const TerminalFormats = {
   Reset: "\x1b[0m",
   Bold: "\x1b[1m",
   Underscore: "\x1b[4m",
   Blink: "\x1b[5m",
   Reverse: "\x1b[7m",
   Hidden: "\x1b[8m",

   FgBlack: "\x1b[30m",
   FgRed: "\x1b[31m",
   FgGreen: "\x1b[32m",
   FgYellow: "\x1b[33m",
   FgBlue: "\x1b[34m",
   FgMagenta: "\x1b[35m",
   FgCyan: "\x1b[36m",
   FgWhite: "\x1b[37m",

   BgBlack: "\x1b[40m",
   BgRed: "\x1b[41m",
   BgGreen: "\x1b[42m",
   BgYellow: "\x1b[43m",
   BgBlue: "\x1b[44m",
   BgMagenta: "\x1b[45m",
   BgCyan: "\x1b[46m",
   BgWhite: "\x1b[47m",
};

export enum IColors {
   NONE,
   RED,
   GREEN,
   YELLOW,
   BLUE,
   MAGENTA,
   CYAN,
   WHITE,
}

export interface IFormatted<T = any> {
   color?: IColors;
   bgcolor?: IColors;
   bold?: boolean;
   italic?: boolean;
   blink?: boolean;
   underscore?: boolean;

   content: T;
}

export class Formatted<T = any> implements IFormatted<T> {
   static strip(formatted: IFormatted<string> | IFormatted<string>[]) {
      if (Array.isArray(formatted)) {
         return formatted.reduce((p, c) => p + c.content, "");
      } else {
         return formatted.content;
      }
   }

   constructor(content?: T, apply?: Formatted) {
      if (content instanceof Formatted) {
         this.color = content.color;
         this.bgcolor = content.bgcolor;
         this.bold = content.bold;
         this.italic = content.italic;
         this.blink = content.blink;
         this.underscore = content.underscore;
         this.content = content.content;
      } else {
         this.content = content;
      }

      if (apply) {
         this.color = apply.color;
         this.bgcolor = apply.bgcolor;
         this.bold = apply.bold;
         this.italic = apply.italic;
         this.blink = apply.blink;
         this.underscore = apply.underscore;
      }
   }

   color?: IColors;
   bgcolor?: IColors;
   bold?: boolean;
   italic?: boolean;
   blink?: boolean;
   underscore?: boolean;

   content: T;

   _color(c: IColors) {
      this.color = c;
      return this;
   }

   _bgcolor(c: IColors) {
      this.bgcolor = c;
      return this;
   }

   _bold(active = true) {
      this.bold = active;
      return this;
   }

   _italic(active = true) {
      this.italic = active;
      return this;
   }

   _blink(active = true) {
      this.blink = active;
      return this;
   }
}

const AddColorFormat = (color: IColors) => (content: any) => {
   if (content instanceof Formatted) {
      content.color = color;
      return content;
   } else {
      const d = new Formatted(content);
      d.color = color;
      return d;
   }
};

const AddBGColorFormat = (color: IColors) => (content: any) => {
   if (content instanceof Formatted) {
      content.bgcolor = color;
      return content;
   } else {
      const d = new Formatted(content);
      d.bgcolor = color;
      return d;
   }
};

export const Format = {
   none: AddColorFormat(IColors.NONE),
   red: AddColorFormat(IColors.RED),
   green: AddColorFormat(IColors.GREEN),
   yellow: AddColorFormat(IColors.YELLOW),
   blue: AddColorFormat(IColors.BLUE),
   magenta: AddColorFormat(IColors.MAGENTA),
   cyan: AddColorFormat(IColors.CYAN),
   white: AddColorFormat(IColors.WHITE),
   bgnone: AddBGColorFormat(IColors.NONE),
   bgred: AddBGColorFormat(IColors.RED),
   bggreen: AddBGColorFormat(IColors.GREEN),
   bgyellow: AddBGColorFormat(IColors.YELLOW),
   bgblue: AddBGColorFormat(IColors.BLUE),
   bgmagenta: AddBGColorFormat(IColors.MAGENTA),
   bgcyan: AddBGColorFormat(IColors.CYAN),
   bgwhite: AddBGColorFormat(IColors.WHITE),
   bold: (content: any) => {
      if (content instanceof Formatted) {
         content.bold = true;
         return content;
      } else {
         const d = new Formatted(content);
         d.bold = true;
         return d;
      }
   },
   italic: (content: any) => {
      if (content instanceof Formatted) {
         content.italic = true;
         return content;
      } else {
         const d = new Formatted(content);
         d.italic = true;
         return d;
      }
   },
};

export interface FormatConfig {
   error: Formatted;
   warning: Formatted;
   log: Formatted;
   debug: Formatted;

   date: Formatted;
   file: Formatted;

   names: Formatted;
   names_delimiter: Formatted;
}

export class DefaultFormatConfig implements FormatConfig {
   error = new Formatted()._color(IColors.RED)._bold();
   warning = new Formatted()._color(IColors.YELLOW)._bold();
   log = new Formatted()._color(IColors.NONE)._bold();
   debug = new Formatted()._color(IColors.CYAN)._bold();

   date = new Formatted()._color(IColors.NONE);
   file = new Formatted()._color(IColors.NONE);
   names = new Formatted()._bold()._color(IColors.CYAN);
   names_delimiter = new Formatted(" -> ")._bold();
}

export interface Message {
   type: LoggingTypes;
   names?: string[];
   text: IFormatted<string>[];
   date: Date;
   file: string;
}

export interface Adapter {
   readonly level: LoggingTypes;

   /**
    * This function initialises the Adapter. It might be called multiple times, when added to multiple instances
    * @param observable An observable to subscribe to messages
    */
   init(): void | Promise<void>;

   flush(sync: true): void;
   flush(sync: false): void | Promise<void>;

   setLevel(level: LoggingTypes): void;

   onMessage(message: Message): void;

   /**
    * When a close function is available, it will be called when no logging instance references it anymore.
    *
    * WARNING: The adapter might be reinitialised, when it is added to a new Logging instance
    */
   close?(): void;
}

export interface ILoggingTimer {
   end: () => number;
}

export interface ILoggingInterface {
   debug(...message: any[]): void;
   log(...message: any[]): void;
   warn(...message: any[]): void;
   error(...message: any[]): void;

   time(name?: string): ILoggingTimer;
   timeEnd(id: string): number;

   getChild(name: string): ILoggingInterface;
}
